#include "base.hpp"

using namespace Engine;
using namespace Application;

/**
*   Private helper methods (application specific)
*/

void Application::AppBase::OnLoaded()
{
    ass->RawBitmapsToBitmaps();
    ass->RawmodelsToModels();

    sceneGame->OnLoaded(ass);
}

void Application::AppBase::DrawToCanvas(Engine::Graphics* g)
{
    g->SetRenderTarget(canvas);
    currentScene->Draw(g);
    g->SetRenderTarget();
}

void Application::AppBase::CalculateCanvasProp()
{
    Point dim = GetGraphics()->GetWinSize();

    if( (float)dim.x/(float)dim.y >= (float)config.canvasSize.x/(float)config.canvasSize.y )
    {
        canvasPos.y = 0.0f;
        canvasPos.x = dim.x/2.0f - ((float)dim.y/(float)config.canvasSize.y)*config.canvasSize.x / 2.0f;
        canvasScale.y = (float)dim.y/(float)config.canvasSize.y;
        canvasScale.x = canvasScale.y;
    }
    else
    {
        canvasPos.x = 0.0f;
        canvasPos.y = dim.y/2.0f - ( (float) dim.x/ (float)config.canvasSize.x)*config.canvasSize.y / 2.0f;
        canvasScale.x = (float)dim.x/(float)config.canvasSize.x;
        canvasScale.y = canvasScale.x;
    }
}

//! --------------------------------- //

/**
* Private methods (overridden)
*/
void Application::AppBase::OnInit()
{
    std::cout << "Application started." << '\n';

    glfwSetInputMode(GetWindow(), GLFW_CURSOR, GLFW_CURSOR_DISABLED);
    ControlState::Init();
    Vpad::Init();

    ass = new Assets();
    assbuf = CreateAssetBuffer(config.assetPath,ass);

    sceneGame = new Game();
    currentScene = sceneGame;

    shader = LoadShader(config.shaderPath + "/" + "vertex.glsl",config.shaderPath + "/" + "fragment.glsl");
    shader->Enable();
    GetGraphics()->InitTransforms(shader);
    GetGraphics()->InitEffects(shader);
    GetGraphics()->SetCanvasSize(config.canvasSize);

    isLoading = false;
    loadingReady = false;
    loadState = 0;

    canvas = new RenderTarget(config.canvasSize.x,config.canvasSize.y);
    CalculateCanvasProp();

    //! @todo Load control config from a file
    Vpad::PushButton(GLFW_KEY_SPACE,0);

    LoadAssets(&assbuf);

}

void Application::AppBase::OnStart()
{
    config = ReadConfigFile("config.cml");
    defWinSize = config.defaultWinSize;
    defFullscreen = config.defaultFullscreenEnabled;
}

void Application::AppBase::OnUpdate(double deltaTime)
{
    timeMod = ((float)deltaTime) / (1.0f/60.0f);

    if(!loadingReady)
    {
        int isLoaded = loadState == 0 ? (int)IsAssetLoadingReady() : true;

        if(loadState == 0)
        {
            if(isLoaded == 1)
            {
                JoinAssetLoaderThread();
                loadingReady = true;
                OnLoaded();
            }
            else if(isLoaded == -1)
            {
                throw std::runtime_error("Asset loading error: " + GetAssetLoaderError());
            }
        }

    }
    else
    {
        Vpad::Update();

        currentScene->Update(timeMod);
    }

    if(ControlState::GetKeyState(GLFW_KEY_LEFT_CONTROL) == Down &&
            ControlState::GetKeyState(GLFW_KEY_Q) == Pressed)
    {
        Terminate();
    }

    if(ControlState::GetKeyState(GLFW_KEY_LEFT_ALT) == Down &&
            ControlState::GetKeyState(GLFW_KEY_ENTER) == Pressed)
    {
        ToggleFulscreen();
    }

    if(ControlState::GetKeyState(GLFW_KEY_LEFT_CONTROL) == Down && ControlState::GetKeyState(GLFW_KEY_F1) == Pressed)
    {
        int mode = glfwGetInputMode(window,GLFW_CURSOR);

        if(mode != GLFW_CURSOR_DISABLED)
            glfwSetInputMode(GetWindow(), GLFW_CURSOR, GLFW_CURSOR_DISABLED);
        else
            glfwSetInputMode(GetWindow(), GLFW_CURSOR, GLFW_CURSOR_NORMAL);
    }

    ControlState::Update();
}

void Application::AppBase::OnResize(int width, int height)
{
    CalculateCanvasProp();
}

void Application::AppBase::OnKeyDown(int key)
{
    ControlState::OnKeyDown(key);
}

void Application::AppBase::OnKeyUp(int key)
{
    ControlState::OnKeyUp(key);
}

void Application::AppBase::OnMouseMove(double x, double y)
{
    ControlState::OnMouseMove(x,y);
}

void Application::AppBase::OnMouseDown(int button)
{
    ControlState::OnMouseDown(button);
}

void Application::AppBase::OnMouseUp(int button)
{
    ControlState::OnMouseUp(button);
}

void Application::AppBase::OnMouseWheel(double offset)
{
    std::cout << offset << '\n';
}

void Application::AppBase::OnDraw(Graphics* g)
{
    if(loadingReady)
    {
        g->Clear(Color(0.0f,0.0f,0.0f));

        DrawToCanvas(g);

        g->Identity(),
        g->Ortho2D(g->GetWinSize().x,g->GetWinSize().y);
        g->UseTransform();

        g->SetLinearFilter(false);
        g->DrawBitmap(canvas,canvasPos.x,canvasPos.y,canvasScale.x,canvasScale.y);
        g->SetLinearFilter(true);
    }
    else
    {

    }
}

void Application::AppBase::OnDestroy()
{
    std::cout << "Application terminated." << '\n';

    JoinAssetLoaderThread();

    sceneGame->Destroy();
}

