#include "application_manager.hpp"

using namespace Engine;

/*
* Private helper/utility functions
*/
void Engine::ApplicationManager::CreateWindow()
{
    window = glfwCreateWindow(640, 480, "Application", nullptr, nullptr);
    if(defFullscreen)
    {
        ToggleFulscreen();
    }
    if(window == nullptr)
    {
        throw std::runtime_error("Failed to open a window.");
    }
    glfwMakeContextCurrent(window);

    int width, height;
    glfwGetWindowSize(window, &width, &height);

    winSize = Point(width,height);

}

// -------------------------------- //

/*
* Protected functions
*/

void Engine::ApplicationManager::OnCloseWindow()
{
    isRunning = false;
}

// -------------------------------- //

/*
* Private functions
*/

void Engine::ApplicationManager::Initialize()
{
    defWinSize = Point(640,480);
    defFullscreen = false;
    OnStart();

    if(!glfwInit())
    {
        throw std::runtime_error("Failed to initialize glfw.");
    }
    InitDisplayHandle();
    CreateWindow();

    if(glewInit() != GLEW_OK)
    {
        throw std::runtime_error("Failed to initialize glew.");
    }

    InitEventHandle(window);

    isRunning = true;

    g = new Graphics();
    g->SetWinSize(winSize);

    oldTime = glfwGetTime();
    newTime = oldTime;

    OnInit();
}

void Engine::ApplicationManager::Update(double deltaTime)
{
    OnUpdate(deltaTime);
}

void Engine::ApplicationManager::Draw(Graphics* g)
{
    glClear(GL_COLOR_BUFFER_BIT);
    glClearColor(0.0f,0.0f,0.0f,1.0f);

    OnDraw(g);

    glfwSwapBuffers(window);
}

void Engine::ApplicationManager::Destroy()
{
    if(programDestroyed) return;

    /*
    * To make sure user-defined variables are destroyed first
    */
    OnDestroy();

    delete g;

    programDestroyed = true;
}

void Engine::ApplicationManager::OnResizeEvent(int width, int height)
{
    glViewport(0,0,width,height);
    winSize = Point(width,height);

    g->SetWinSize(winSize);

    OnResize(width,height);

}

// -------------------------------- //

/*
* Public functions
*/

void Engine::ApplicationManager::Run(int argc, char ** argv)
{
    programDestroyed = false;

    try
    {
        Initialize();

        while(isRunning)
        {
            if(glfwWindowShouldClose(window))
            {
                OnCloseWindow();
            }
            /*
            * Update game logic and draw the scene
            */

            newTime = glfwGetTime();
            if(newTime-oldTime >= 1.0d/60.0d)
            {
                Update(newTime-oldTime);
                oldTime = newTime;

                Draw(g);
            }

            glfwPollEvents();
        }

        Destroy();
    }
    catch(std::runtime_error e)
    {
        // Handles application specific exceptions
        std::cout << e.what() << '\n';

        // Destroy vital components now
        Destroy();

        // Throws an exception, so the user-defined exception handle may catch it
        throw std::runtime_error("An error occurred, application terminated.");

    }
}

Engine::ApplicationManager::~ApplicationManager()
{
    Destroy();
}
