#include "render_utility.hpp"

using namespace Engine;

void Engine::RenderUtility::DrawMesh(Engine::Mesh* m, Bitmap* bmp)
{
    if(m == nullptr) return;
    m->Draw(bmp,linFilter);
}

void Engine::RenderUtility::DrawMesh(Engine::Mesh* m, GLuint tex)
{
    if(m == nullptr) return;
    m->Draw(tex,linFilter);
}

void Engine::RenderUtility::DrawModel(Engine::Model* m, Bitmap* bmp)
{
    if(m == nullptr) return;
    m->Draw(bmp,linFilter);
}

void Engine::RenderUtility::DrawModel(Engine::Model* m, std::vector<Bitmap*> bitmaps)
{
    if(m == nullptr) return;
    m->Draw(bitmaps,linFilter);
}

void Engine::RenderUtility::Clear(Color c)
{
    glClearDepth(1.0f);
    glClearColor(c.r,c.g,c.b,c.a);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

void Engine::RenderUtility::SetDepthTest(bool mode)
{
    if(mode)
        glEnable(GL_DEPTH_TEST);
    else
        glDisable(GL_DEPTH_TEST);
}

void Engine::RenderUtility::SetBlending(bool mode)
{
    if(mode)
    {
        glEnable (GL_BLEND);
        glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    }
    else
    {
        glDisable(GL_BLEND);
    }
}

void Engine::RenderUtility::DrawText(Bitmap* font, std::string text, float x, float y)
{
    PushMatrix();
    Translate(x,y,0.0f);
    Scale(font->GetSize().x / 16.0f,font->GetSize().y / 16.0f,1.0f);
    UseTransform();

    unsigned char c;

    int column;
    int row;

    for(unsigned int i = 0; i < text.length(); i++)
    {
        c = text[i];

        row = floor(c/16);
        column = c -row*16;

        fontPart->Draw(font,column,row,linFilter);

        Translate( (font->GetSize().x / 16.0f / 1.75f) / (font->GetSize().x/16.0f),0.0f,0.0f);
        UseTransform();
    }

    PopMatrix();
}

void Engine::RenderUtility::DrawBitmap(Bitmap* img, float x, float y, float scalex, float scaley)
{
    PushMatrix();

    Translate(x,y,0.0f);
    Scale(img->GetSize().x * scalex, img->GetSize().y*scaley, 1.0f);
    UseTransform();

    DrawMesh(mBmpFull,img);

    PopMatrix();
}

void Engine::RenderUtility::DrawBitmap(RenderTarget* rt, float x, float y, float scalex, float scaley)
{
    PushMatrix();

    Translate(x,y,0.0f);
    Scale(rt->GetSize().x * scalex, rt->GetSize().y*scaley, 1.0f);
    UseTransform();

    DrawMesh(mBmpFullFlipY,rt->GetTexture());

    PopMatrix();
}

void Engine::RenderUtility::DrawBitmapRegion(Bitmap* img, float sx, float sy, float sw, float sh, float dx, float dy, float dw, float dh)
{
    PushMatrix();

    Translate(dx,dy,0.0f);
    Scale(dw, dh, 1.0f);
    UseTransform();

    float iw = img->GetSize().x;
    float ih = img->GetSize().y;

    sx /= iw;
    sy /= ih;
    sw /= iw;
    sh /= ih;

    mBmpReg->NewUVCoords(std::vector<GLfloat>
    {
        sx,sy,
        sx+sw,sy,
        sx+sw,sy+sh,
        sx,sy+sh,
    });

    DrawMesh(mBmpReg,img);

    PopMatrix();
}

void Engine::RenderUtility::DrawTriangle(float x1, float y1, float x2, float y2, float x3, float y3)
{
    bool v = SetTexturesEnabled(false);

    mTriangle->NewVertices(std::vector<GLfloat>
    {
        x1,y1,0.0f,
        x2,y2,0.0f,
        x3,y3,0.0f,
    });

    mTriangle->Draw(nullptr);

    SetTexturesEnabled(v);
}

void Engine::RenderUtility::SetRenderTarget(RenderTarget* rt)
{
    if(rt != nullptr)
    {
        rt->UseFrameBuffer();
    }
    else
    {
        glBindFramebuffer(GL_FRAMEBUFFER, 0);
        glViewport(0,0,winSize.x,winSize.y);
    }
}
void Engine::RenderUtility::SetLinearFilter(bool mode)
{
    linFilter = mode;
}

void Engine::RenderUtility::InitRenderUtility()
{
    fontPart = GetPartition(Point(16,16));

    mBmpFull = new Mesh(std::vector<GLfloat>{0.0f,0.0f,0.0f, 1.0f,0.0f,0.0f, 1.0f,1.0f,0.0f, 0.0f,1.0f,0.0f},
        std::vector<GLuint>{0,1,2, 2,3,0},
        std::vector<GLfloat>{0.0f,0.0f, 1.0f,0.0f, 1.0f,1.0f, 0.0f,1.0f},
        std::vector<GLfloat>{0.0f,0.0f,1.0f, 0.0f,0.0f,1.0f, 0.0f,0.0f,1.0f, 0.0f,0.0f,1.0f});

    mBmpFullFlipY = new Mesh(std::vector<GLfloat>{0.0f,0.0f,0.0f, 1.0f,0.0f,0.0f, 1.0f,1.0f,0.0f, 0.0f,1.0f,0.0f},
        std::vector<GLuint>{0,1,2, 2,3,0},
        std::vector<GLfloat>{0.0f,1.0f, 1.0f,1.0f, 1.0f,0.0f, 0.0f,0.0f},
        std::vector<GLfloat>{0.0f,0.0f,1.0f, 0.0f,0.0f,1.0f, 0.0f,0.0f,1.0f, 0.0f,0.0f,1.0f});

    mBmpReg = new Mesh(std::vector<GLfloat>{0.0f,0.0f,0.0f, 1.0f,0.0f,0.0f, 1.0f,1.0f,0.0f, 0.0f,1.0f,0.0f},
        std::vector<GLuint>{0,1,2, 2,3,0},
        std::vector<GLfloat>{0.0f,0.0f, 1.0f,0.0f, 1.0f,1.0f, 0.0f,1.0f},
        std::vector<GLfloat>{0.0f,0.0f,1.0f, 0.0f,0.0f,1.0f, 0.0f,0.0f,1.0f, 0.0f,0.0f,1.0f});

    mTriangle = new Mesh(std::vector<GLfloat>{0.0f,0.0f,0.0f, 1.0f,0.0f,0.0f, 1.0f,1.0f,0.0f},
        std::vector<GLuint>{0,1,2},
        std::vector<GLfloat>{0.0f,0.0f, 1.0f,0.0f, 1.0f,1.0f},
        std::vector<GLfloat>{0.0f,0.0f,1.0f, 0.0f,0.0f,1.0f, 0.0f,0.0f,1.0f, 0.0f,0.0f,1.0f});

    linFilter = true;
}
