#include "shader.hpp"

using namespace Engine;

/*
    TODO: Add helper functions to reduce the length of the following function
    Or maybe not
*/
Engine::Shader::Shader(std::string vertexPath,std::string fragPath)
{

 	// Create the shaders
 	GLuint vertexShaderID = glCreateShader(GL_VERTEX_SHADER);
 	GLuint fragmentShaderID = glCreateShader(GL_FRAGMENT_SHADER);

 	// Read vertex shader
 	std::string vertexShaderCode;
 	std::ifstream vertexShaderStream(vertexPath.c_str(), std::ios::in);

 	if(vertexShaderStream.is_open())
 	{
 		std::string line = "";
 		while(getline(vertexShaderStream, line))
 		{
			vertexShaderCode += "\n" + line;
 		}
 		vertexShaderStream.close();
 	}
 	else
 	{
 		throw std::runtime_error("Failed to open vertex shader!");
 	}

 	// Read fragment shader
 	std::string fragmentShaderCode;
 	std::ifstream fragmentShaderStream(fragPath.c_str(), std::ios::in);

 	if(fragmentShaderStream.is_open())
 	{
 		std::string line = "";
 		while(getline(fragmentShaderStream, line))
 		{
 			fragmentShaderCode += "\n" + line;
 		}
 		fragmentShaderStream.close();
 	}
 	else
 	{
        throw std::runtime_error("Failed to open fragment shader!");
 	}

 	GLint result = GL_FALSE;
 	int infoLogLength;

 	// Compile vertex shader
 	char const * VertexSourcePointer = vertexShaderCode.c_str();
 	glShaderSource(vertexShaderID, 1, &VertexSourcePointer , nullptr);
 	glCompileShader(vertexShaderID);

 	// Check Vertex Shader
 	glGetShaderiv(vertexShaderID, GL_COMPILE_STATUS, &result);
 	glGetShaderiv(vertexShaderID, GL_INFO_LOG_LENGTH, &infoLogLength);

    std::string errors = "";
 	if (infoLogLength > 0)
 	{
 		std::vector<char> VertexShaderErrorMessage(infoLogLength+1);
 		glGetShaderInfoLog(vertexShaderID, infoLogLength, nullptr, &VertexShaderErrorMessage[0]);

        if(std::string(&VertexShaderErrorMessage[0]).length() > 0 )
        {
            errors += std::string(&VertexShaderErrorMessage[0] ) + ":";
 		}

        if(errors.length() > 0)
        {
            throw std::runtime_error( ( (std::string)("Vertex shader error: ") + errors).c_str() );
        }
 	}

 	// Compile Fragment Shader
 	char const * FragmentSourcePointer = fragmentShaderCode.c_str();
 	glShaderSource(fragmentShaderID, 1, &FragmentSourcePointer , nullptr);
 	glCompileShader(fragmentShaderID);

 	// Check Fragment Shader
 	glGetShaderiv(fragmentShaderID, GL_COMPILE_STATUS, &result);
 	glGetShaderiv(fragmentShaderID, GL_INFO_LOG_LENGTH, &infoLogLength);

    errors = "";
 	if (infoLogLength > 0)
 	{
 		std::vector<char> FragmentShaderErrorMessage(infoLogLength+1);
 		glGetShaderInfoLog(fragmentShaderID, infoLogLength, nullptr, &FragmentShaderErrorMessage[0]);

 		if(std::string(&FragmentShaderErrorMessage[0]).length() > 0 )
        {
            errors += std::string(&FragmentShaderErrorMessage[0] ) + ":";
 		}

        if(errors.length() > 0)
        {
            throw std::runtime_error(  ( (std::string)("Fragment shader error: ") + errors).c_str() );
        }

 	}

 	// Link the program
 	GLuint programID = glCreateProgram();

 	glBindAttribLocation(programID,0,"vertexPos");
    glBindAttribLocation(programID,1,"vertexUV");
    glBindAttribLocation(programID,2,"normalCoords");

 	glAttachShader(programID, vertexShaderID);
 	glAttachShader(programID, fragmentShaderID);

 	glLinkProgram(programID);

 	// Check the program
 	glGetProgramiv(programID, GL_LINK_STATUS, &result);
 	glGetProgramiv(programID, GL_INFO_LOG_LENGTH, &infoLogLength);

 	errors = "";
 	if ( infoLogLength > 0 )
 	{
 		std::vector<char> ProgramErrorMessage(infoLogLength+1);
 		glGetProgramInfoLog(programID, infoLogLength, nullptr, &ProgramErrorMessage[0]);

 		if(std::string(&ProgramErrorMessage[0]).length() > 0 )
        {
            errors += std::string(&ProgramErrorMessage[0] ) + ":";
 		}

        if(errors.length() > 0)
        {
            throw std::runtime_error( ( (std::string)("Shader error: ") + errors).c_str() );
        }
 	}

 	glDetachShader(programID, vertexShaderID);
 	glDetachShader(programID, fragmentShaderID);

 	glDeleteShader(vertexShaderID);
 	glDeleteShader(fragmentShaderID);

 	if(programID == 0)
 	{
 		throw std::runtime_error("Failed to load shaders! Shader program returned zero.");
 	}
 	std::cout << "Shader program loaded successfully." << '\n';

 	shaderProgram = programID;
}

void Engine::Shader::Enable()
{
    glUseProgram(shaderProgram);
}

void Engine::Shader::Disable()
{
    glUseProgram(0);
}

GLuint Engine::Shader::GetProgram()
{
    return shaderProgram;
}

Engine::Shader * Engine::LoadShader(std::string vertexPath,std::string fragPath)
{
    return new Shader(vertexPath.c_str(),fragPath.c_str());
}

Engine::Shader::~Shader()
{
    glDeleteProgram(shaderProgram);
}
