#include "transform.hpp"

using namespace Engine;

/**
* Public class methods
*/
void Engine::Transform::InitTransforms(Shader* sh)
{
    GLuint program = sh->GetProgram();

    model = glm::mat4(1.0f);
    projection = glm::mat4(1.0f);
    view = glm::mat4(1.0f);

    currentMatrix = 0;

    if(modelStorage.size() == 0) modelStorage = std::vector<glm::mat4> ();
    if(viewStorage.size() == 0)  viewStorage = std::vector<glm::mat4> ();
    if(projectionStorage.size() == 0) projectionStorage = std::vector<glm::mat4> ();

    modelID = program != 0 ? glGetUniformLocation(program, "model") : 0;
    viewID = program != 0 ? glGetUniformLocation(program, "view") : 0;
    projectionID = program != 0 ? glGetUniformLocation(program, "projection") : 0;
}

void Engine::Transform::View(glm::vec3 position, glm::vec3 direction, glm::vec3 up)
{
    view = glm::lookAt(position,direction,up);
}

void Engine::Transform::View(glm::mat4 newView)
{
    view = newView;
}

void Engine::Transform::Projection(float FOV, Point frame, float _min, float _max)
{
    projection = glm::perspective(FOV * (float)M_PI / 180.0f, (float)frame.x/(float)frame.y, _min, _max);
}

void Engine::Transform::Identity()
{
    model = glm::mat4();
}

void Engine::Transform::Scale(float x,float y,float z)
{
    model *= glm::scale(glm::mat4(1.0f),glm::vec3(x,y,z) );
}

void Engine::Transform::Translate(float x,float y, float z)
{
    model *= glm::translate(glm::mat4(1.0f),glm::vec3(x,y,z) );
}

void Engine::Transform::Rotate(float angle,glm::vec3 axis)
{
    model *= glm::rotate(glm::mat4(1.0f),angle,axis);
}

void Engine::Transform::RotateMultiple(std::vector<RotateData> rotations)
{
    glm::mat4 rm = glm::mat4(1.0f);
    for(auto r : rotations)
    {
        rm *= glm::rotate(glm::mat4(1.0f),r.angle,r.axes);
    }

    model *= rm;
}

void Engine::Transform::Ortho2D(float width, float height)
{
    projection = glm::ortho(0.0f, width, height, 0.0f);
    view = glm::mat4(1.0f);
}

void Engine::Transform::PushMatrix()
{
    currentMatrix++;
    if(currentMatrix > (int)modelStorage.size())
    {
        modelStorage.push_back(glm::mat4(1.0f));
        projectionStorage.push_back(glm::mat4(1.0f));
        viewStorage.push_back(glm::mat4(1.0f));
    }

    modelStorage[currentMatrix-1] = glm::mat4(model);
    projectionStorage[currentMatrix-1] = glm::mat4(projection);
    viewStorage[currentMatrix-1] = glm::mat4(view);

}

void Engine::Transform::PopMatrix()
{
    currentMatrix--;

    model = glm::mat4(modelStorage[currentMatrix]);
    projection = glm::mat4(projectionStorage[currentMatrix]);
    view = glm::mat4(viewStorage[currentMatrix]);
}

void Engine::Transform::ResetBufferIndex()
{
    currentMatrix = 0;
}

void Engine::Transform::UseTransform()
{
    glUniformMatrix4fv(modelID, 1, GL_FALSE, &model[0][0]);
    glUniformMatrix4fv(viewID, 1, GL_FALSE, &view[0][0]);
    glUniformMatrix4fv(projectionID, 1, GL_FALSE, &projection[0][0]);
}
