#include "tmxparser.hpp"

using namespace TmxParser;
using namespace rapidxml;

/**
    TmxMap class function definitions
**/

TmxParser::TmxMap::TmxMap()
{
    width = 0;
    height = 0;
}

TmxParser::TmxMap::TmxMap(unsigned int nw, unsigned int nh)
{
    width = nw;
    height = nh;
}

void TmxParser::TmxMap::PushLayer(TmxLayer l)
{
    layers.push_back(l);
}

void TmxParser::TmxMap::PushProperty(KeyValuePair prop)
{
    properties.push_back(prop);
}

unsigned int TmxParser::TmxMap::GetTile(unsigned int layer, unsigned int x,unsigned int y)
{
    // Check if (x,y) is out of range
    if(layer > layers.size() || x > width-1 || y > height-1) return 0;

    return layers[layer].data[y*width+x];
}

std::string TmxParser::TmxMap::GetPropertyValue(std::string name)
{
    for(auto p : properties)
    {
        if(p.key == name)
        {
            return p.value;
        }
    }

    return "";
}

unsigned int TmxParser::TmxMap::GetLayerCount()
{
    return layers.size();
}

// -------------------------------------------- //


/**
    Static variables
**/
static std::string err;
static bool errTriggered;
static xml_document<> * doc;
static TmxMap* mpointer;


/**
    Static helper functions
**/

static std::vector<unsigned int> RawDataToArray(std::string raw)
{
    std::vector<unsigned int> data;

    unsigned int startPos = 0;
    unsigned int value = 0;

    for(unsigned int i = 0; i < raw.length(); i++)
    {
        if(raw[i] == ',')
        {
            // To make sure there is something between commas
            if(raw.length()-startPos > 0)
            {
                value =  (unsigned int) std::stoi( raw.substr(startPos,raw.length()-startPos) ) ;
                data.push_back(value);
            }
            startPos = i+1;


        }
    }

    value =  (unsigned int) std::stoi( raw.substr(startPos,raw.length()-startPos) ) ;
                data.push_back(value);

    return data;
}

static std::string LoadDocument(std::string path)
{
    std::ifstream file { path };

    if(file.is_open() == false)
    {
        errTriggered = true;
        err = "Failed to open file in " + path;
        return "";
    }

    std::string content { std::istreambuf_iterator<char>(file), std::istreambuf_iterator<char>() };

    return content;
}

static int ParseSubnodes(xml_node<> *n)
{
    std::string name;

    for (xml_node<> *node = n->first_node(); node; node = node->next_sibling())
    {
        name = std::string(node->name());

        if(name == "properties")
        {
            /*
                Go through properties and store 'em
            */
            for (xml_node<> *nn = node->first_node(); nn; nn = nn->next_sibling())
            {
                if(std::string(nn->name()) == "property")
                {
                    KeyValuePair kvp;
                    for (xml_attribute<> *attr = nn->first_attribute(); attr; attr = attr->next_attribute())
                    {
                        if(std::string(attr->name()) == "name")
                        {
                            kvp.key = doc->allocate_string(attr->value());
                        }
                        else if(std::string(attr->name()) == "value")
                        {
                            kvp.value = doc->allocate_string(attr->value());
                        }
                    }
                    mpointer->PushProperty(kvp);
                }
            }
        }
        else if(name == "layer")
        {
            TmxLayer layer;

            for (xml_attribute<> *attr = node->first_attribute(); attr; attr = attr->next_attribute())
            {
                if(std::string (attr->name()) == "name")
                {
                    layer.name = doc->allocate_string(attr->value()) ;
                }
            }

            for (xml_node<> *nn = node->first_node(); nn; nn = nn->next_sibling())
            {
                if(std::string(nn->name()) == "data")
                {
                    for (xml_attribute<> *attr = nn->first_attribute(); attr; attr = attr->next_attribute())
                    {
                        if(std::string (attr->name()) == "encoding")
                        {
                            if(std::string (attr->value()) != "csv")
                            {
                                err = "Error in layer '" + layer.name + "': the encoding must be 'csv'!";
                                return 1;
                            }
                        }
                    }
                    std::string raw = doc->allocate_string(nn->value()) ;
                    layer.data = RawDataToArray(raw);
                    mpointer->PushLayer(layer);
                }
            }
        }
    }

    return 0;
}

static int ParseNodes()
{
    for (xml_node<> *node = doc->first_node(); node; node = node->next_sibling())
    {
        for (xml_attribute<> *attr = node->first_attribute(); attr; attr = attr->next_attribute())
        {
            std::string name = (attr->name());

            if(name == "width")
            {
                mpointer->SetWidth(std::stoi(doc->allocate_string(attr->value()) ));
            }
            else if(name == "height")
            {
                mpointer->SetHeight(std::stoi( doc->allocate_string(attr->value())));
            }
        }

        if (ParseSubnodes(node) == 1)
            return 1;
    }

    return 0;
}

/**
    Outside-class functions
**/

TmxMap* TmxParser::LoadTmxMap(std::string path)
{
    TmxMap* m = new TmxMap();
    mpointer = m;
    doc = new xml_document<> ;
    err = "";
    errTriggered = false;

    std::string data = LoadDocument(path);
    if(errTriggered)
    {
        return nullptr;
    }

    doc->parse<0>((char*)data.c_str());
    if(ParseNodes() == 1)
        return nullptr;

    /*
        To make sure it's empty when we next time load a tilemap

        ...Maybe not required.
    */
    delete doc;

    return m;
}

std::string TmxParser::GetError()
{
    return err;
}
